import logging
import os
import pathlib

from image_service_foundation.config.configuration_manager import ConfigurationManager
from image_service_foundation.storage.dropbox_cloud import DropBoxStorage
from image_service_foundation.storage.google_cloud import GoogleStorage

logger = logging.getLogger(__name__)


class ModelsProvider:
    def __init__(self, config: ConfigurationManager, models_converter):
        self.config = config
        self.storage_client_name = self.config.get('storage.client', 'dropbox')
        self._init_client(self.storage_client_name)
        self.enabled_models = self.config.get(
            'models.enabledTypes'
        )  # TODO: move if not os.path.isdir(self.save_path): to model_converter.save_tf()
        self.model_converter = models_converter

        if not self.enabled_models:
            raise ValueError(
                "No models enabled in configuration. "
                "Specify a list of enabled model types with key `models.enabledTypes`."
            )
        self.service_name = self.config.get('storage.serviceName')
        if not self.service_name:
            raise ValueError(
                "No service name specified in configuration. Specify the service name with key storage.serviceName"
            )

        self.destination_folder = self.config.get('models.location', f"/etc/{self.service_name}/models")

    def _init_client(self, storage_client='dropbox'):
        if storage_client == 'gcloud':
            self.storage_client = GoogleStorage(anonymous=True)
        elif storage_client == 'dropbox':
            self.storage_client = DropBoxStorage(self.config['dropbox.accesstoken'])
        else:
            raise NotImplementedError('Currently only supported for gcloud & dropbox')
        logger.info(f"Initialized ModelsProvider with storage client {storage_client}")

    def _get_file_destination(self, model_type, filename):
        destination_folder = os.path.join(self.destination_folder, model_type)
        destination_path = os.path.join(destination_folder, os.path.basename(filename))
        return destination_path

    def download_models(self):
        for model_type in self.enabled_models:
            logger.info(f"Attempting download of {model_type}")
            try:
                blob_name = self.config.data[f'{model_type}.modelName']
            except KeyError:
                raise ValueError(f"No configuration value found for `{model_type}.modelName`.")

            file_destination = self._get_file_destination(model_type, blob_name)
            folder_destination = os.path.dirname(os.path.abspath(file_destination))
            os.makedirs(folder_destination, exist_ok=True)

            file_destination = pathlib.Path(file_destination)
            if not file_destination.exists() or file_destination.stat().st_size == 0:
                bucket_name = self.config.get('storage.bucket', 'md-data-bingemarkers')
                self.storage_client.download(bucket_name, blob_name, file_destination)

            self.model_converter.save_tf()



import datetime
import pathlib
import typing

from google.cloud.storage import Blob, Bucket, Client


class BlobNotFoundException(FileNotFoundError):
    pass


class GoogleStorage:
    def __init__(
        self, service_account_filepath: typing.Union[str, pathlib.Path] = None, anonymous: bool = False
    ) -> None:
        if anonymous:
            self.client = Client.create_anonymous_client()
        else:
            self.client = (
                Client.from_service_account_json(service_account_filepath) if service_account_filepath else Client()
            )

    def download(
        self, bucket_name: str, blob_name: typing.Union[str, pathlib.Path], filepath: typing.Union[str, pathlib.Path]
    ) -> None:
        bucket = Bucket(self.client, name=bucket_name)
        blob = bucket.get_blob(str(blob_name))
        if blob is None:
            raise BlobNotFoundException(f"Could not find blob with name {blob_name} in bucket {bucket_name}")
        blob.download_to_filename(str(filepath))

    def upload(
        self,
        bucket_name: str,
        destination_blob_name: typing.Union[str, pathlib.Path],
        filepath: typing.Union[str, pathlib.Path],
        content_type: str = None,
    ) -> None:
        bucket = Bucket(self.client, name=bucket_name)
        blob = bucket.blob(destination_blob_name)
        blob.upload_from_filename(str(filepath), content_type=content_type)

    def list(self, bucket_name: str) -> typing.Iterator[Blob]:
        return self.client.list_blobs(bucket_name)

    def generate_download_signed_url_v4(self, bucket_name, blob_name, ttl_days=7):
        """Generates a v4 signed URL for downloading a blob.

        Note that this method requires a service account key file. You can not use
        this if you are using Application Default Credentials from Google Compute
        Engine or from the Google Cloud SDK.
        """

        bucket = Bucket(self.client, name=bucket_name)
        blob = bucket.blob(blob_name)

        url = blob.generate_signed_url(version="v4", expiration=datetime.timedelta(days=ttl_days), method="GET")

        return url



import pathlib
import typing

import dropbox


class DropBoxStorage:
    def __init__(self, accesstoken: str) -> None:
        self.accesstoken = accesstoken

    def download(
        self, bucket_name: str, blob_name: typing.Union[str, pathlib.Path], filepath: typing.Union[str, pathlib.Path]
    ) -> None:
        connection = dropbox.Dropbox(self.accesstoken)
        with connection:
            with open(filepath, 'wb') as f:
                metadata, result = connection.files_download(path=str(blob_name))
                f.write(result.content)



import logging
import os
import pathlib
import typing

logger = logging.getLogger(__name__)

PathLike = typing.Union[str, pathlib.Path]


class DefaultSpringCloudConfigParser:
    def _extract_name(self, name: str) -> str:
        basename = os.path.basename(name)
        return os.path.splitext(basename)[0]

    def __call__(self, raw_data: dict, *args, **kwargs) -> dict:
        expected = raw_data['name']
        for source in raw_data['propertySources']:
            if source['name'].endswith('application.yaml'):
                config = source['source']
                break
        else:
            logger.warning('No application.yaml found in response; using empty config as base')
            config = {}

        names = []
        for source in raw_data['propertySources']:
            if source['name'].endswith('application.yaml'):
                continue
            else:
                name = self._extract_name(source['name'])
                names.append(name)
                config.update(source['source'])
        if expected not in names:
            raise ValueError("expected config with name '{}' was not found in response".format(expected))

        return config



import copy
import json
import logging
import pathlib
import re

import requests
import yaml

from image_service_foundation.config.utils import (
    DefaultSpringCloudConfigParser,
    PathLike,
)

logger = logging.getLogger(__name__)


class ConfigurationMixin:
    """
    Configuration class mixin. Should be called once during initialization.
    """

    def load_config(self, config=None, default_location=None):
        logger.debug('Loading ConfigurationManager')
        if config is not None:
            logger.debug('from passed argument')
            self._config = config
        else:
            if default_location is None:
                raise ValueError('default location is required if no configuration instance is provided')
            logger.debug('from default location at {}'.format(default_location))
            self._config = ConfigurationManager.from_path(default_location)

    @property
    def config(self):
        return self._config


def flatten_dict(data):
    def expand(key, value):
        if isinstance(value, dict):
            return [(key + '.' + k, v) for k, v in flatten_dict(value).items()]
        else:
            return [(key, value)]

    items = [item for k, v in data.items() for item in expand(k, v)]

    return dict(items)


def reindex_list(data):
    pattern = '\[\d+\]'  # looking for instances like hello[0], goodbye[1]
    incorrect_indexed_items = [k for k, v in data.items() if re.search(pattern, k)]
    for item in incorrect_indexed_items:
        correct_name = re.sub(pattern, '', item)
        if correct_name in data:
            data[correct_name].append(data[item])
        else:
            data[correct_name] = [data[item]]
    return data


class ConfigurationManager:
    """
    Configuration class that contains a dictionary with all configuration options as loaded from a YAML/JSON formatted
    file or remote Spring Cloud Config URL.

    :param should_reindex_list: = Boolean to indicate whether to iterate through the configurations to load correctly
    for spring cloud.
    """

    def __init__(self, data: dict, location: str, flatten=False, should_reindex_list=True):
        self._data = copy.deepcopy(data)
        if flatten:
            self._data = flatten_dict(self._data)
        if should_reindex_list:
            self._data = reindex_list(self._data)
        self._location = str(location)

    @property
    def data(self):
        return self._data

    @property
    def location(self):
        return self._location

    def __contains__(self, item):
        return item in self.data

    def __getitem__(self, item):
        return self.data[item]

    def get(self, item, default=None):
        return self.data.get(item, default)

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__, self.data, self.location)

    def __str__(self):
        s = '{}('.format(self.__class__.__name__)
        s += '{},\n'.format(json.dumps(self.data, indent=2, sort_keys=True))
        s += '{!r}\n)'.format(self.location)
        return s

    def refresh(self):
        self._data = ConfigurationManager.from_path(self.location).data

    def _validate_path(self, filepath: PathLike) -> pathlib.Path:
        if filepath is None:
            if self.location.startswith(('http://', 'https://')):
                raise ValueError('filepath argument is required if the configuration was retrieved from remote')
            else:
                return pathlib.Path(self.location)
        else:
            return pathlib.Path(filepath)

    def to_file(self, filepath: PathLike = None, **kwargs):
        filepath = self._validate_path(filepath)
        if filepath.suffix.lower().endswith(('yaml', 'yml')):
            self.to_yaml_file(filepath, **kwargs)
        elif filepath.suffix.lower().endswith('json'):
            self.to_json_file(filepath, **kwargs)
        else:
            raise NotImplementedError("extension '{}' is not supported".format(filepath.suffix))

    def to_json_file(self, filepath: PathLike = None, **_):
        filepath = self._validate_path(filepath)
        with open(str(filepath), 'w') as f:
            json.dump(self.data, f)

    def to_yaml_file(self, filepath: PathLike = None, *, safe=True):
        filepath = self._validate_path(filepath)
        with open(str(filepath), 'w') as f:
            yaml.dump(self.data, f, yaml.SafeDumper if safe else yaml.Dumper)

    @classmethod
    def from_path(cls, filepath_or_url, **kwargs) -> 'ConfigurationManager':
        if str(filepath_or_url).startswith(('http://', 'https://')):
            return cls.from_remote(filepath_or_url, **kwargs)
        else:
            return cls.from_file(filepath_or_url, **kwargs)

    @classmethod
    def from_file(cls, filepath: PathLike, **kwargs) -> 'ConfigurationManager':
        filepath = pathlib.Path(filepath)
        if filepath.suffix.lower().endswith(('yaml', 'yml')):
            return cls.from_yaml_file(filepath, **kwargs)
        elif filepath.suffix.lower().endswith('json'):
            return cls.from_json_file(filepath, **kwargs)
        else:
            raise NotImplementedError("extension '{}' is not supported".format(filepath.suffix))

    @classmethod
    def from_yaml_file(cls, filepath: PathLike, *, safe=True, **kwargs) -> 'ConfigurationManager':
        with open(str(filepath), 'r') as f:
            data = yaml.load(f, Loader=yaml.SafeLoader if safe else yaml.FullLoader)
            if data is None:
                raise IOError('failed to load YAML file; is the file empty?')
            return cls(data, str(filepath), **kwargs)

    @classmethod
    def from_json_file(cls, filepath: PathLike, **kwargs) -> 'ConfigurationManager':
        with open(str(filepath), 'r') as f:
            data = json.load(f)
            return cls(data, str(filepath), **kwargs)

    @classmethod
    def from_remote(
        cls, url: str, *, parser: callable = DefaultSpringCloudConfigParser(), **kwargs
    ) -> 'ConfigurationManager':
        response = requests.get(url)
        if not response.ok:
            response.raise_for_status()
        raw = response.json()
        config = parser(raw)
        return cls(config, str(url), **kwargs)



import logging
import sys


def set_logging(filepath=None, stream=None, level=logging.INFO, multithreaded=False, multiprocessed=False):
    """Only top level scripts should call this function."""
    log_format = '%(asctime)s - '
    if multiprocessed:
        log_format += '%(process)d - '
    log_format += '%(levelname)-8s - '
    if multithreaded:
        log_format += '%(threadName)s - '
    log_format += '%(name)s - %(message)s'

    level = level.upper() if isinstance(level, str) else level

    if not (filepath is None or stream is None):
        raise ValueError("'filepath' and 'stream' should not be specified together")
    elif filepath is not None:
        logging.basicConfig(level=level, format=log_format, filename=filepath)
    elif stream is not None:
        logging.basicConfig(level=level, format=log_format, stream=stream)
    else:
        logging.basicConfig(level=level, format=log_format)

    sys.excepthook = _exception_hook


def _exception_hook(exc_type, exc_value, exc_traceback):
    """Simple function to ensure unexpected exceptions are still logged"""
    logging.exception('Uncaught exception', exc_info=(exc_type, exc_value, exc_traceback))
